/********************************************/
/*    ACアダプタテスト回路                  */
/********************************************/
//    外部A/D MCP3204  12bit 4ch 2.5V入力
//    外部D/A MCP4821  12bit 1ch 2.048V出力
//    2014-03-04    開始
//
//    PB0    o  D/A CS
//    PB1    o  D/A SCK
//    PB2    o　D/A SDI
//    PB3    o　CC on/off
//    PB5    o　P-P reset on/off
//    PD2    o  A/D CLK
//    PD3    i  A/D DOUT
//    PD4    o  A/D DIN
//    PD5    o  A/D CS
//    PD6    i  CC Err Hで定電流制御エラー
//    ﾀｲﾏｰ1  1mSｼｽﾃﾑﾀｲﾏｰ
//    ﾀｲﾏｰ2  160uS A/D変換ﾀｲﾏｰ
//    外部A/D ch0 直流電圧
//            ch1 ピーク値入力
//            ch2 -
//            ch3 -
//    内部A/D ch0 AC100V 交流電圧 100Vで1V
//            ch1 AC100V 交流電流 CT 1Aで0.5V
//    Vref = 2.5V (外付け)

/*****  ライブラリ      ****/
#include  "wiring_private.h"      // sbi,cbi命令用
#include  "extern.h"              // structのためにめに

/*****  設定値      *****/
//    外付けA/D平均回数  1サイクル=4ms
#define  AD_AVRS    75       // 300msで確定
//    内蔵A/D平均回数  1サイクル=0.16ms
#define  ADI_AVRS   1875     // 300msで確定
//    回路により誤差が生じるので実測して設定
//    外付けA/D ch0 : 電圧測定スケーリングデータ
struct STCAL1 vt_scl = {
    3383,  2500,     // X1(A/D),Y1(x.xxV)
};
//    外付けA/D ch1 : P-P値スケーリングデータ
struct STCAL2 pp_scl = {
     35,    4,      // X1(A/D),Y1(x.xxV)
    4060,  489,     // X2,Y2
};
//    内蔵A/D ch0 : RMS 電圧スケーリングデータ
struct STCALF1 rmsv_scl = {
    287.0,  100.0,      // X1(A/D),Y1(xxxV)
};
//    内蔵A/D ch1 : RMS 電流スケーリングデータ
struct STCALF1 rmsa_scl = {
    230.0,  0.60,     // X1(A/D),Y1(x.xxA)
};
//    電力(W)スケーリングデータ
struct STCALF1 wat_scl = {
    66158.5,  60.00,    // X1(A/D),Y1(xx.xxW)
};
//    D/A : 定電流値スケーリングデータ
struct STCAL2 da_scl = {
     127,  100,     // X1(mA),Y1(D/A)
    5214, 4000,     // X2,Y2
};

//  内部 A/D オフセットデータ
//  読み出したA/D値に加算してゼロ点補正する
int adi_offset[] = {
      4,            // ch0 電圧入力
      3,            // ch1 電流入力
};

/*****  ピン制御マクロ    *****/
#define  ADCLK_H   (sbi(PORTD,PORTD2))
#define  ADCLK_L   (cbi(PORTD,PORTD2))
#define  ADDIN_H   (sbi(PORTD,PORTD4))
#define  ADDIN_L   (cbi(PORTD,PORTD4))
#define  ADCS_H    (sbi(PORTD,PORTD5))
#define  ADCS_L    (cbi(PORTD,PORTD5))
#define  DACS_H    (sbi(PORTB,PORTB0))
#define  DACS_L    (cbi(PORTB,PORTB0))
#define  DASCK_H   (sbi(PORTB,PORTB1))
#define  DASCK_L   (cbi(PORTB,PORTB1))
#define  DASDI_H   (sbi(PORTB,PORTB2))
#define  DASDI_L   (cbi(PORTB,PORTB2))
#define  CC_ON     (sbi(PORTB,PORTB3))
#define  CC_OFF    (cbi(PORTB,PORTB3))
#define  PRST_ON   (cbi(PORTB,PORTB5))
#define  PRST_OFF  (sbi(PORTB,PORTB5))
#define  ADDO_INP  (PIND & (1 << PORTD3))
#define  CCER_INP  (PIND & (1 << PORTD6))
#define  NOP       asm("nop")
//    テストポイント
#define  PC2_H     (sbi(PORTC,PORTC2))    // (!!!) PC2 25pin ﾄｸﾞﾙ
#define  PC2_L     (cbi(PORTC,PORTC2))
#define  PC3_X     (sbi(PINC,PINC3))      // (!!!) PC3 26pin ﾄｸﾞﾙ
#define  PC3_H     (sbi(PORTC,PORTC3))
#define  PC3_L     (cbi(PORTC,PORTC3))
#define  PC4_H     (sbi(PORTC,PORTC4))    // (!!!) PC4 27pin ﾄｸﾞﾙ
#define  PC4_L     (cbi(PORTC,PORTC4))
#define  PC5_H     (sbi(PORTC,PORTC5))    // (!!!) PC5 28pin ﾄｸﾞﾙ
#define  PC5_L     (cbi(PORTC,PORTC5))

/*****  単純な時間待ち    *****/
//    100で50uS弱
void wait(volatile word n)
{
    while(n){
        n--;
    }
}

/*****************************/
/*    スケーリング           */
/*****************************/

/*****  1点校正でのスケーリング  *****/
//      in  d　入力値
//          s  cal1データ
//             x:入力と同じ単位の基準値
//             y:出力と同じ単位の基準値
//      out word  答
word scale1(word d, STCAL1 s)
{
float a;
//  PC4_H;                       // (!!!) PC4 27pin ﾄｸﾞﾙ
    a = ((float)d * (float)s.y) / (float)s.x;
    a += 0.5;                    // 正で四捨五入
    if(a < 0.0)    a= 0.0;       // マイナスはゼロに
//  PC4_L;                       // (!!!)
    return (word)a;
}

/*****  2点校正でのスケーリング  *****/
//      in  d　入力値
//          s  cal2データ
//             x1,x2:入力と同じ単位の基準値
//             y1,y2:出力と同じ単位の基準値
//      out word  答
word scale2(word d, STCAL2 s)
{
float a;
//  PC4_H;                       // (!!!) PC4 27pin ﾄｸﾞﾙ
    a = ((((float)s.y2 - (float)s.y1)  /      // Y
          ((float)s.x2 - (float)s.x1)) *      // X
          ((float)d - (float)s.x1)) +
           (float)s.y1;
    a += 0.5;                    // 正で四捨五入
    if(a < 0.0)    a= 0.0;       // マイナスはゼロに
//  PC4_L;                       // (!!!)
    return (word)a;
}

/*****  floatでの1点校正でのスケーリング  *****/
//      in  d　入力値
//          s  cal1データ
//             x:入力と同じ単位の基準値
//             y:出力と同じ単位の基準値
//      out float  答
float scalef1(float d, STCALF1 s)
{
float a;
//  PC4_H;                       // (!!!) PC4 27pin ﾄｸﾞﾙ
    a = (d * s.y) / s.x;
//  PC4_L;                       // (!!!)
    return a;
}

/*****  2点校正でのスケーリング  *****/
//      in  d　入力値
//          s  cal2データ
//             x1,x2:入力と同じ単位の基準値
//             y1,y2:出力と同じ単位の基準値
//      out word  答
float scalef2(float d, STCALF2 s)
{
float a;
//  PC4_H;                       // (!!!) PC4 27pin ﾄｸﾞﾙ
    a = (((s.y2 - s.y1)  /  (s.x2 - s.x1)) *    // Y / X
          (d - s.x1)) + s.y1;
    if(a < 0.0)    a= 0.0;       // マイナスはゼロに
//  PC4_L;                       // (!!!)
    return a;
}

/****************************/
/*    外付けA/D入力         */
/****************************/
//    ch数は4
//    1msごとに変換　　4msで4ch完了
//    ch0:電圧  ch1:P-P値   ch2,3:未使用
volatile word ad_data[4];     // A/D 生データ 12bit
volatile word ad_avr[4];      // A/D 平均値 (adokで確定)
volatile long ad_add[4];      // 平均加算値
volatile word ad_pkck[4];     // A/D ピーク比較値
volatile word ad_peak[4];     // A/D ピークデータ (adokで確定)
volatile byte ad_ch;          // 変換チャンネル 0～3
volatile byte ad_cnt;         // 平均加算回数  75回:300ms
volatile boolean f_adok;      // A/D 変換完了フラグﾞ 

/*****  A/D入力      *****/
//      MCP3204を制御
//      12bit値を返す
//      およそ25usで変換
word adread(byte ch)
{
byte i;
word d = 0;
    ADCS_H;         // A/D CS  = H
    ADCLK_L;        // A/D CLK = L
    ADCS_L;         // A/D CS  = L
//  変換ch指定    
    ch |= B11000;          // スタートビットとシングルエンドビットを付加
    ch <<= 3;                      // 計5ビットを送出
    for(i = 0; i < 7; i++){        // clkパルスは7つ
      if(ch & 0x80)   ADDIN_H;     // MSBから出力
      else            ADDIN_L;
      ADCLK_H;                     // A/D CLK = H
      ch <<= 1;                    // 左シフト
      NOP;  NOP;  NOP;             // wait
      ADCLK_L;                     // A/D CLK = L
    }
    ADDIN_H;                       // A/D DIN = H
//  A/D変換完了データ読み取り
    for(i = 0; i < 12; i++){       // clkパルス
      d <<= 1;
      if(ADDO_INP)      d |= 1;    // on
      NOP;
      ADCLK_H;                     // A/D CLK = H
      NOP;  NOP;  NOP;
      ADCLK_L;                     // A/D CLK = L
    }
    ADCS_H;         // A/D CS  = H
    ADCLK_H;        // A/D CLK = H
    return d;
}

/*****  A/D 平均処理      *****/
//    1msに1回ｺｰﾙ
//    4chを順次変換
//    平均回数75回なので300msで平均値確定
void adavr(void)
{
word d;
byte i;
//  A/D変換
    d = adread(ad_ch);        // 12bit A/D
    ad_data[ad_ch] = d;       // 生データ保存
    ad_add[ad_ch] += (long)d; // 加算
//  ピーク値
    if(d > ad_pkck[ad_ch]){    // ピーク保持値を超えたとき
        ad_pkck[ad_ch] = d;    // そのデータに
    }
//  次ch
    ad_ch++;
    if(ad_ch >= 4){           // 4ch完了
      ad_ch = 0;
      ad_cnt++;               // 平均回数
      if(ad_cnt >= AD_AVRS){  // 75回目で平均処理
        PC4_H;                // (!!!) PC4 27pin ﾄｸﾞﾙ
        for(i = 0; i< 4; i++){                 // 4ch
          ad_avr[i] = (word)(ad_add[i] / AD_AVRS);  // 平均
          ad_add[i] = 0;             // 加算値をゼロに
          ad_peak[i] = ad_pkck[i];   // ピーク値確定
          ad_pkck[i] = ad_data[i];   // 現在値をセット
        }
        ad_cnt = 0;           // 回数を0に        
        f_adok = 1;           // データ確定フラグをオン
        PC4_L;                // (!!!)
      }
    }
} 

/*****  A/D入力電圧計算  *****/
//      in  A/D入力ﾃﾞｰﾀ 12bit
//      out 0.01V単位の電圧値
word advolt(word d)
{
    return scale1(d, vt_scl);
}

/*****  A/Dピーク値電圧計算  *****/
//      in  A/D入力ﾃﾞｰﾀ 12bit
//      out 0.01V単位の電圧値
int adpeak(word d)
{
    return scale2(d, pp_scl);
}

/******************************/
/*      内部Ａ／Ｄ処理        */
/******************************/

/****  A/D データ        *****/
//    AD0  交流電圧入力
//    AD1  交流電流入力
volatile byte  adi_ch;          // A/D変換チャンネル
                                // 0,1 : min,max処理
//  AD0,AD1 データ (電圧,電流)
volatile int   adi_data[2];     // A/D変換データ +/-512
                                // 0:電圧入力,1:電流
volatile long  adi_add[3];      // 加算合計
                                // 0:電圧,1:電流,2:電力
volatile long  adi_sum[3];      // 加算結果 addをコピー
                                // 0:電圧,1:電流,2:電力
volatile word  adi_cnt;         // サンプル回数
                                // 1000回
volatile boolean f_adiok;       // サンプル完了フラグ 
volatile boolean f_adi2ch;      // A/Dデータ2ch分取得完了フラグ
                                // リセットタイミング待ちで使用
//  内蔵A/D 電圧,電流,電力 計算結果
float adi_avrv;                 // ch0 電圧平均値 sum/cnt
float adi_avra;                 // ch1 電流平均値
float adi_avrp;                 // ch0 * ch1 電力平均値
float adi_rmsv;                 // 電圧RMS(V)
float adi_rmsa;                 // 電流RMS(A)
float adi_pwr;                  // 実効電力（W)
float adi_va;                   // 皮相電力（VA)

//    A/D サンプリング実験
#define  ADI_SMPL    140        // サンプル数
volatile int   adi_smplv[ADI_SMPL];     // 電圧
volatile int   adi_smpla[ADI_SMPL];     // 電流
volatile byte  adi_scnt;         // カウント
volatile boolean f_adismpl;      // サンプル指令フラグ 

/*****  A/D 変換チャンネル指定    *****/
byte adi_mux[]={
    0b00000000,       // ch0 入力1
    0b00000001,       // ch1 入力2
//    ||| ++++----  MUX ch
//    ||+---------  ADLAR
//    ++----------  REFS   AREFピン (2.5V)
};

/*****  A/D割り込み処理    *****/
//    タイマー割り込みでA/D変換開始
//    ch0：電圧 ch1：電流入力
ISR(ADC_vect)
{
static int d;
    PC5_H;                      // (!!!) PC5 28pin ﾄｸﾞﾙ
//  次ch 変換  (少しでもスピードアップ)
    if(adi_ch == 0){
        ADMUX = adi_mux[1];     // 次ch(ch1)指定
        ADCSRA |= (1 << ADSC);  // A/D変換開始
    }
//  現データリード  ch0,1ch1
    d = (word)ADCL;             // A/D
    d |= (ADCH & 0x03) << 8;    // 符号なしで 0～3FF
    d -= 512;                   // +/-512に
//  測定値
    d += adi_offset[adi_ch];         // ゼロ点オフセット
    adi_data[adi_ch] = d;            // データを保存
//  電圧,電流値を二乗して加算
    adi_add[adi_ch] += (long)d * (long)d;  //RMS計算用
//  次変換ch
    adi_ch++;                         // 次ch
    if(adi_ch >= sizeof(adi_mux)){    // 2ch終了?
        f_adi2ch = 1;                 // フラグで知らせる
//  電圧*電流で瞬時電力
        adi_add[2] += (long)adi_data[0] * (long)d; 
//  サンプル回数チェック
        adi_cnt++;                    // 回数+1
        if(adi_cnt >= ADI_AVRS){      // 1875回?
            adi_sum[0] = adi_add[0];  // 電圧
            adi_sum[1] = adi_add[1];  // 電流
            adi_sum[2] = adi_add[2];  // 電力
            adi_add[0] = adi_add[1] = adi_add[2] = 0;
            adi_cnt = 0;              // 回数クリア
            f_adiok = 1;              // 結果確定
        }
//  サンプリング実験
        if(f_adismpl){                // サンプリング実験
            adi_smplv[adi_scnt] = adi_data[0];    // 電圧
            adi_smpla[adi_scnt] = adi_data[1];    // 電流
            adi_scnt++;
            if(adi_scnt >= ADI_SMPL){    // 終わり?
                f_adismpl = 0;           // フラグをクリア
            }
        }
    }
    PC5_L;                    // (!!!)
}

/*****  内部A/D 電圧,電流,電力 測定値計算  *****/
void adiscale(void)
{
long sum[3];        // 電圧、電流、電力
    cli();                          // 割り込み禁止
    sum[0] = adi_sum[0];            // 電圧
    sum[1] = adi_sum[1];            // 電流
    sum[2] = adi_sum[2];            // 電力
    sei();                          // 割り込み再開
//  RMS値 電力計算
    adi_avrv = sqrt((float)sum[0] / (float)ADI_AVRS);  // 電圧平均値
    adi_avra = sqrt((float)sum[1] / (float)ADI_AVRS);  // 電流平均値
    adi_avrp = (float)sum[2] / (float)ADI_AVRS;        // 電力平均値
    adi_rmsv = scalef1(adi_avrv, rmsv_scl);            // 電圧RMS(V)
    adi_rmsa = scalef1(adi_avra, rmsa_scl);            // 電流RMS(A)
    adi_pwr  = scalef1(adi_avrp, wat_scl);             // 実効電力（W)
    adi_va   = adi_rmsv * adi_rmsa;                    // 皮相電力（VA)
}

/*****  A/D 平均,加算処理リセット    *****/
void adreset(void)
{
//  ピーク検出リセット
    PRST_ON;                   // ピークホルダリセットオン
    wait(200);                 // wait
    PRST_OFF;                  // ピークホールド開始
    wait(200);                 // wait
//  A/D割り込み2chの完了を待つ
    f_adi2ch = 0;              // フラグをクリアして
    while(f_adi2ch == 0);      // 割り込みで1になるのを待つ
//  割り込み禁止にしてクリア
    PC3_H;                    // (!!!) PC3 26pin ﾄｸﾞﾙ
    cli();                    // 割り込み禁止
//  外付けA/D 4ch
    ad_add[0]  = ad_add[1]  = ad_add[2]  = ad_add[3]  = 0; // 加算値をゼロに
    ad_pkck[0] = ad_pkck[1] = ad_pkck[2] = ad_pkck[3] = 0; // ピーク値もゼロに
    ad_cnt = 0;               // 回数を0に        
    ad_ch  = 0;               // ch = 0
    f_adok = 0;               // データ確定フラグをオフ
//  内蔵A/D 電圧,電流,電力
    adi_add[0] = adi_add[1] = adi_add[2] = 0;  // 加算値をゼロに
    adi_cnt = 0;              // 回数
    adi_ch  = 0;              // ch = 0
    f_adiok = 0;              // データ確定フラグをオフ
    sei();                    // 割り込み再開
    PC3_L;
}

/******************************/
/*      タイマー割り込み      */
/******************************/

/*****  タイマー値    *****/
volatile boolean f_tm1ms;    // 1msありﾌﾗｸﾞ 
volatile byte  tm_1ms;       // 1msﾀﾞｳﾝｶｳﾝﾄﾀｲﾏｰ max 255ms

/*****  タイマー０：1kHz割り込み    *****/
ISR(TIMER0_COMPA_vect)
{
//  PC4_H;                       // (!!!) PC4 27pin ﾄｸﾞﾙ
//  1ms計時処理（ダウンカウント）
    f_tm1ms = 1;
    if(tm_1ms)      tm_1ms--;
//  PC4_L;
}

/*****  タイマー２：6.25kHz割り込み    *****/
//      A/Dに対して変換指令
ISR(TIMER2_COMPA_vect)
{
    PC5_H;              // (!!!) PC5 28pin ﾄｸﾞﾙ
//  A/D変換　データの読み出しは変換完了割り込みで
    adi_ch = 0;                  // ch 0から
    ADMUX = adi_mux[0];          // ch指定
    ADCSRA |= (1 << ADSC);       // A/D変換開始
    PC5_L;                       // (!!!)test
}

/******************************/
/*      定電流ｏｎ／ｏｆｆ    */
/******************************/
boolean  cc_cont;              // CC on/off

/*****  定電流on/off出力  *****/
//    cc_contでｵﾝ,ｵﾌを
void cccont(void)
{
    if(cc_cont)    CC_ON;
    else           CC_OFF;
}

/*****  定電流ｵﾝ      *****/
void ccon(void)
{
    CC_ON;
    cc_cont = 1;
}

/*****  定電流ｵﾌ      *****/
void ccoff(void)
{
    CC_OFF;
    cc_cont = 0;
}

/*****  定電流on/offトグル  *****/
void cctoggle(void)
{
    cc_cont ^= 1;      // トグル
    cccont();          // on/off
}

/*****  定電流状態出力    *****/
void txcccont(void)
{
    Serial.print("# CC ");  
    if(cc_cont == 0)    Serial.println("off");  
    else                Serial.println("on");  
}

/************************************/
/*      外部D/A出力処理             */
/************************************/

/*****  12bit D/A出力       *****/
/*      MCP4921を制御           */
/*      0～4095                 */
void daout(word d)
{
byte i;
    d |= 0b0111000000000000;
    //     ||||++++++++++++--- data
    //     |||+--------------- /SHDN 
    //     ||+---------------- /GA  x1
    //     |+----------------- BUF
    //     +------------------ /A
    DASCK_L;                       // SCK L
    DACS_L;                        // CS L
    for(i = 0; i < 16; i++){       // 16bit loop
        if(d & 0x8000){            // bit15をチェック
            DASDI_H;               // data H
        }
        else{
            DASDI_L;               // data L
        }
        DASCK_H;                   // CLK H
        d <<= 1;                   // 左シフト
        DASCK_L;                   // CLK L
    }
    DACS_H;                        // CS H
    DASCK_H;                       // CLK H
    DASDI_H;                       // data L
}

/****  D/A出力電流計算    *****/
//    in  電流値 xxxx mA
//    out D/A設定 12bitデータ 0～4095
word daamp(word d)
{
    return scale2(d, da_scl);
}

/******************************/
/*      シリアル受信処理      */
/******************************/
#define  RXBF_SIZ    10      // ﾊﾞｯﾌｧ文字数
char rx_bff[RXBF_SIZ+1];     // 受信ﾊﾞｯﾌｧ
boolean f_rxok;              // 受信完了ﾌﾗｸﾞ

/*****  ｼﾘｱﾙ1行受信      *****/
void rxbff(void)
{
static byte cnt = 0;            // 受信文字数  
char c;  
    if(Serial.available()){    // ﾃﾞｰﾀあり
      if(f_rxok == 0){          // 前ﾃﾞｰﾀ受信処理した
        c = Serial.read();
        if(c == '\r'){            // CR?
            rx_bff[cnt] = '\0';   // NULLを最後に
            f_rxok = 1;           // 受信成功
            cnt = 0;              // 最初から
        }
        else{                     // 文字
          if((cnt < RXBF_SIZ) &&  // ﾊﾞｯﾌｧサイズok
             (isprint(c))){       // 20～7E
            rx_bff[cnt] = c;      // ﾊﾞｯﾌｧにｽﾄｱ
            cnt++;                // 1文字進める
          }
        }
      }
    }
}

/*****  xxxxx シリアルデータ出力    *****/
void tx6d(int d)
{
char s[16];
    sprintf(s, "%6d", d);
    Serial.print(s);
}

/*****  xxxxx.x シリアルデータ出力    *****/
//    intを1/10してdp付加して出力
void tx5r1(int d)
{
char s[16];
double a;
    a = (double)d / 10.0;         // 1/10
    dtostrf(a, 7, 1, s);          // ｢%f｣が使えないので
    Serial.print(s);
}

/*****  xxxx.xx シリアルデータ出力    *****/
//    intを1/100してdp付加して出力
void tx4r2(int d)
{
char s[16];
double a;
    a = (double)d / 100.0;        // 1/100
    dtostrf(a, 7, 2, s);          // ｢%f｣が使えないので
    Serial.print(s);
}

/*****  xxx.xxx シリアルデータ出力    *****/
//    intを1/1000してdp付加して出力
void tx3r3(int d)
{
char s[16];
double a;
    a = (double)d / 1000.0;       // 1/1000
    dtostrf(a, 7, 3, s);          // ｢%f｣が使えないので
    Serial.print(s);
}

/*****  xxxxx.x シリアルデータ出力    *****/
//    floatを出力
void tx5f1(float f)
{
char s[16];
    dtostrf(f, 7, 1, s);         // ｢%f｣が使えないので
    Serial.print(s);
}

/*****  xxxxx.x シリアルデータ出力    *****/
//    floatを出力
void tx4f2(float f)
{
char s[16];
    dtostrf(f, 7, 2, s);         // ｢%f｣が使えないので
    Serial.print(s);
}

/*****  xxxxx.x シリアルデータ出力    *****/
//    floatを出力
void tx8f2(float f)
{
char s[16];
    dtostrf(f, 11, 2, s);         // ｢%f｣が使えないので
    Serial.print(s);
}

/****************************/
/*        実行              */
/****************************/
//    電流電圧測定データ
word  cur_start;            // 測定開始電流
word  cur_end;              // 測定終了電流
word  cur_step;             // 電流増加値
                            // 1mA単位
byte  minp_cnt;             // 測定パラメータ入力実行番号
                            // 0:開始電流、1:終了電流、2:電流増加
const char *msr1_msg[] = {
    "# Cur Start",
    "# Cur End  ",
    "# Cur Step ",
};
word *msr1_dp[] = {
      &cur_start,
      &cur_end,
      &cur_step,
};
word  msr_cur;              // 測定電流
                            // 0.000A～5.000A
//    実行区分
byte exc_j = 0;           // 実行区分

#define J_TST      1      // テストモード選択
#define J_TSTAD    3      // A/Dテスト
#define	J_TSTDA    4      // D/Aテスト
#define	J_TSTCUR   6      // CURテスト
#define	J_MSR1     8      // 0-5000mA測定データ入力
#define	J_MSRMSG   10     // 測定メッセージ
#define	J_MSRLP    11     // 0-10A測定loop
#define J_MSRBK    14     // 中断
#define J_TSTSMPL  15     // 内蔵A/D電圧,電流サンプルテスト
#define J_TSTIAD   19     // 内蔵A/Dテスト RMS電圧,電流,電力
#define J_MON      20     // モニター 外付A/Dと内蔵A/Dのスケーリング値

/*****  タイトル出力      *****/
void jttl(void)
{
    Serial.println("# AC adaptert tester (2014-05-01)");
    exc_j++;
}

/*****  テストモード選択      *****/
void jtst(void)
{
    Serial.print("\n# 0:0-5A A:A/D I:int-A/D M:Mon D:D/A C:Cur S:sample>");
    f_rxok = 0;
    exc_j++;
}

/*****  テストモード受信待ち      *****/
void jtstrx(void)
{
char c;
    if(f_rxok){                // 受信あり
        f_rxok = 0;
        Serial.println();
        c = toupper(rx_bff[0]);    // 先頭文字
        switch(c){
          case '0':
            Serial.println("# 0-5000mA Test (1000=1A)");
            minp_cnt = 0;          // 数値入力区分
            exc_j = J_MSR1;
            break;
          case '1':
            Serial.println("# 0-5A, 10mA step");
            cur_start = 0;           // 測定開始電流
            cur_end   = 5000;        // 測定終了電流
            cur_step  = 10;          // 電流増加値
            exc_j = J_MSRMSG;
            break;
          case '2':
            Serial.println("# 0-5A, 20mA step");
            cur_start = 0;           // 測定開始電流
            cur_end   = 5000;        // 測定終了電流
            cur_step  = 20;          // 電流増加値
            exc_j = J_MSRMSG;
            break;
          case '5':
            Serial.println("# 0-5A, 50mA step");
            cur_start = 0;           // 測定開始電流
            cur_end   = 5000;        // 測定終了電流
            cur_step  = 50;          // 電流増加値
            exc_j = J_MSRMSG;
            break;
          case 'A':
            Serial.println("# A/D test");
            adreset();              // A/D処理リセット
            exc_j = J_TSTAD;
            break;
          case 'D':
            Serial.println("# D/A test");
            exc_j = J_TSTDA;
            break;
          case 'C':
            Serial.println("# Cur test");
            exc_j = J_TSTCUR;
            break;
          case 'S':
            Serial.println("# A/D sample test");
            exc_j = J_TSTSMPL;
            break;
          case 'I':
            Serial.println("# int A/D test");
            adreset();              // A/D処理リセット
            exc_j = J_TSTIAD;
            break;
          case 'M':
            Serial.println("# Monitor");
            adreset();              // A/D処理リセット
            exc_j = J_MON;
            break;
        default:
            Serial.print("# ?");
            exc_j = J_TST;
        }
    }
}

/*****  A/Dテスト    *****/
//    測定完了タイミングに1行 外付けA/Dデータを表示
//    2chだけの数値
void jtstad(void)
{
word d;
    if(f_adok){
        f_adok = 0;
//  ch0 直流電圧
        d = ad_avr[0];              // 外付A/D ch0 入力電圧
        tx6d(d);                    // BCD出力
        tx4r2(advolt(d));           // xxxx.xx       
        Serial.print("V");
//  ch1 ピーク電圧
        d = ad_avr[1];              // 外付A/D ch1 入力電圧
        tx6d(d);                    // BCD出力
        d = ad_peak[1];             // 外付A/D ch1 ピーク
        tx6d(d);                    // BCD出力
        tx4r2(adpeak(d));           // xxxx.xx       
        Serial.print("V(p-p)");
//  CC Err入力
        if(CCER_INP)    Serial.println(" C-Er");
        else            Serial.println(" C-Ok");
//  リセット
        adreset();                  // A/D処理リセット
    }
    chktstend();        // おわりチェック
}

/*****  A/D テストおわりチェック  *****/
//    入力文字なしで終了
//    数字入力で定電流値設定
//    数字以外で定電流on/off
void chktstend(void)
{
int d;
    if(f_rxok){                     // 受信ﾁｪｯｸ
      f_rxok = 0;
      if(rx_bff[0] == '\0')    exc_j = J_TST; // おわり
      else  if(!isdigit(rx_bff[0])){   // 先頭文字数字以外
          cctoggle();    // 定電流on/offトグル
          txcccont();    // on/off状態出力
      }
      else{
          d = atoi(rx_bff);
          Serial.print("# D/A=");
          Serial.print(d, DEC);
          Serial.println(" mA");
          d = daamp(d);            // mAからD/A値に
          daout(d);
      }
    }
}

/*****  D/Aテスト    *****/
void jtstda(void)
{
    Serial.print("# D/A 0-4095, Ent:end, a:CC >");
    exc_j++;
}

/*****  D/Aテスト実行    *****/
//      12bitでD/A出力
void jtstda1(void)
{
word d;
    if(f_rxok){                  // 受信ﾁｪｯｸ
      f_rxok = 0;
      if(rx_bff[0] == '\0'){
          exc_j = J_TST;         // おわり
      }
      else  if(!isdigit(rx_bff[0])){   // 先頭文字数字以外
          cctoggle();                  // 定電流on/offトグル
          Serial.println();
          txcccont();                  // on/off状態出力
          exc_j--;
      }
      else{
          d = atoi(rx_bff);
          Serial.println(d, DEC);
          daout(d);
          exc_j--;
      }
    }
}

/*****  CURテスト    *****/
//    電流値をmA単位で入力
void jtstcur(void)
{
    Serial.print("# Cur 0-5000mA, Ent:end, a:CC >");
    exc_j++;
}

/*****  CURテスト実行    *****/
//    Enterだけでおわり
//    数字で電流値
//    文字で定電流on/off
void jtstcur1(void)
{
word d;
    if(f_rxok){                  // 受信ﾁｪｯｸ
      f_rxok = 0;
      if(rx_bff[0] == '\0'){
          exc_j = J_TST;         // おわり
      }
      else  if(!isdigit(rx_bff[0])){   // 先頭文字数字以外
          cctoggle();                  // 定電流on/offトグル
          Serial.println();
          txcccont();                  // on/off状態出力
          exc_j--;
      }
      else{
          d = atoi(rx_bff);
          d = daamp(d);            // mAからD/A値に
          Serial.print("# D/A=");
          Serial.println(d, DEC);
          daout(d);
          exc_j--;
      }
    }
}

/*****  0～5000mA 測定    *****/
//      開始電流、終了電流、電流増加値 入力項目出力
void jmsr1(void)
{
    Serial.print(msr1_msg[minp_cnt]);    // 入力項目
    Serial.print(" >");
    exc_j++;
}

/*****  0～5000mA 負荷測定    *****/
//      開始電流、終了電流、電流増加値入力
void jmsr2(void)
{
word d;
word *p;
    if(f_rxok){                  // 受信ﾁｪｯｸ
      f_rxok = 0;
      if(! isdigit(rx_bff[0])){       // 先頭文字数字以外
          Serial.println("-Break");
          exc_j = J_TST;              // 入力中断
      }
      else{
          d = atoi(rx_bff);         // 入力データ
          p = msr1_dp[minp_cnt];    // 設定するデータのアドレス
          *p = d;                   // 入力データストア
          Serial.println(d, DEC);  // 数値を表示
          minp_cnt++;               // 次データ
          if(minp_cnt >= 3){        // データは3つ  3つで測定開始
              exc_j = J_MSRMSG;
          }
          else{
              exc_j--;                    // 入力残り繰り返し
          }
      }
   }
}

/*****  測定開始メッセージ   *****/
void jmsrmsg(void)
{
    msr_cur = cur_start;        // 定電流出力開始
    daout(daamp(msr_cur));      // mAからD/A値に
    ccon();                     // 定電流on
    Serial.println("#  CC A   DC V  P-P V   DC W  AC VA   AC W DC/AC%");
    tm_1ms = 200;               // 200mS待ち
    exc_j++;
}

/*****  測定開始   *****/
//      定電流出力 loop
void jmsrlp(void)
{
    if(tm_1ms == 0){               // タイムアップ  
        daout(daamp(msr_cur));     // mAからD/A値に
        ccon();                    // 定電流on
        tm_1ms = 100;              // 100mS待ち
        exc_j++;
    }
}

/*****  定電流安定待ち    ****/
void jmsrlp1(void)
{
    if(tm_1ms == 0){              // タイムアップ
        adreset();                // A/D平均処理リスタート
        exc_j++;
    }
}   
   
/*****  測定値出力       ****/
void jmsrlp2(void)
{
word volt, pp;      // 入力電圧,P-P値,一次電流
float dcw, ef;      // DC Wと効率
    if(CCER_INP){                  // 定電流制御できない
        Serial.println("# CC Err");
        exc_j = J_MSRBK;           // おわり
    }
    if(f_rxok){                    // 受信ﾁｪｯｸ
        f_rxok = 0;
        Serial.println("# Break");
        exc_j = J_MSRBK;           // おわり
    }
// A/D測定値出力
    if(f_adok && f_adiok){           // 外付け+内蔵A/D平均処理完了
        adiscale();                  // 内蔵A/D 電圧,電流,電力 測定値計算
        volt = advolt(ad_avr[0]);    // 直流入力電圧 0.01 V
        pp   = adpeak(ad_peak[1]);   // P-P電圧 0.01 V
        dcw  = ((float)msr_cur / 1000.0) *   // DC電力
               ((float)volt / 100.0);        // 負荷電流 * 電圧
        if(adi_pwr <= 0.0)       ef = 0.0;            // 効率
        else if(dcw >= adi_pwr)  ef = 1.0;            // 100%
        else                     ef = dcw / adi_pwr;  // DC W / 実効電力
//  結果出力
        tx3r3(msr_cur);              // 定電流値 xxx.xxx mA
        tx4r2(volt);                 // 直流入力電圧 xxxx.xx V
        tx4r2(pp);                   // P-P電圧  xxxx.xx V
        tx4f2(dcw);                  // DC電力   xxxx.xx W
        tx4f2(adi_va);               // 皮相電力 xxxx.xx VA
        tx4f2(adi_pwr);              // 実効電力 xxxx.xx W
        tx5f1(ef * 100.0);           // 効率    xxxxx.x %
        Serial.println();
//  次データ
        msr_cur += cur_step;         // 電流増加
        if(msr_cur > cur_end){       // おわり?
            exc_j = J_MSRBK;
        }
        else{                        // まだ
            exc_j = J_MSRLP;         // loop
        }
    }
}

/*****  測定 中断    *****/
void jmsrbk(void)
{
    ccoff();            // 定電流ｵﾌ
    daout(0);
    exc_j = J_TST;      // ﾃｽﾄﾓｰﾄﾞ選択へ
}

/*****  電圧,電流A/Dサンプリングテスト  *****/
//      電流値を入力してから実行
void jsmpl1(void)
{
    Serial.print("# Cur 0-5000mA, Ent:end >");
    exc_j++;
}

/*****  電圧,電流A/Dサンプリングテスト  *****/
//      電流値を入力
void jsmpl2(void)
{
word d;
    if(f_rxok){                  // 受信ﾁｪｯｸ
      f_rxok = 0;
      if(rx_bff[0] == '\0'){
          exc_j = J_TST;         // おわり
      }
      else{
          d = atoi(rx_bff);
          Serial.println();
          Serial.print("#");
          Serial.print(d, DEC);
          Serial.println("mA");
          daout(daamp(d));           // mAからD/A値に
          ccon();                    // 定電流on
          tm_1ms = 250;              // 250mS待ち
          exc_j++;
      }
    }
}

/*****  電圧,電流A/Dサンプリングテスト  *****/
//      タイムアップで開始
void jsmpl3(void)
{
    if(tm_1ms == 0){
        cli();              // 割り込み禁止
        adi_scnt  = 0;      // 最初から
        f_adismpl = 1;      // サンプリング開始
        sei();              // 割り込み再開
        exc_j++;            // next
    }
}

/*****  電圧,電流A/Dサンプリングテスト  *****/
//      結果出力
void jsmpl4(void)
{
byte i;
    if(f_adismpl == 0){      // サンプリング完了
        ccoff();             // 定電流ｵﾌ
        daout(0);            // 電流ゼロに
        for(i = 0; i < ADI_SMPL; i++){
            tx6d(i);
            tx6d(adi_smplv[i]);     // 電圧
            tx6d(adi_smpla[i]);     // 電流
            Serial.println();
        }
        exc_j = J_TST;       // ﾃｽﾄﾓｰﾄﾞ選択へ
    }
}

/*****  内蔵A/Dテスト    *****/
//    300msごとに1行 RMS電圧,電流,電力を表示
void jtstiad(void)
{
int d0,d1;
    if(f_adiok){              // データ確定
        f_adiok = 0;
        adiscale();           // 内部A/D 電圧,電流,電力 測定値計算
//  内蔵A/D
        cli();                      // いったん割り込み禁止
        d0 = adi_data[0];           // ch0 電圧 A/D ±512
        d1 = adi_data[1];           // ch1 電流 A/D
        sei();
        tx6d(d0);                   // 電圧 生A/D値
        tx6d(d1);                   // 電流 生A/D値
//  ch0 電圧
        tx5f1(adi_avrv);
        tx4f2(adi_rmsv);
        Serial.print("V");
//  ch1 電流
        tx5f1(adi_avra);
        tx4f2(adi_rmsa);
        Serial.print("A");
//  VA,W
        tx4f2(adi_va);
        Serial.print("VA");
        tx8f2(adi_avrp);
        tx4f2(adi_pwr);
        Serial.println("W");
//  リセット
        adreset();                  // A/D処理リセット
    }
    chktstend();        // おわりチェック
}

/*****  モニター表示    *****/
//    外付A/Dと内蔵A/Dの測定値を順次表示
void jmon(void)
{
    if(f_adok && f_adiok){           // 外付け+内蔵A/D平均処理完了
        adiscale();                  // 内蔵A/D 電圧,電流,電力 測定値計算
        tx4r2(advolt(ad_avr[0]));    // 直流入力電圧 xxxx.xx V
        Serial.print("V(DC)");
        tx4r2(adpeak(ad_peak[1]));    // P-P電圧  xxxx.xx V
        Serial.print("V(p-p)");
//  ch0 電圧
        tx4f2(adi_rmsv);
        Serial.print("V(rms)");
//  ch1 電流
        tx4f2(adi_rmsa);
        Serial.print("A(rms)");
//  VA,W
        tx4f2(adi_va);
        Serial.print("VA");
        tx4f2(adi_pwr);
        Serial.println("W");
//  リセット
        adreset();                   // A/D平均処理リスタート
    }
    chktstend();        // おわりチェック
}

/*****  実行テーブル    *****/
void (*tblexc[])(void)={
        jttl,           //    0  タイトル出力
        jtst,           // *  1  テストモード選択
        jtstrx,         //    2  テストモード受信待ち
	jtstad,		// *  3  A/Dテスト
	jtstda, 	// *  4  D/Aテスト
	jtstda1, 	//    5  D/Aテスト 実行
	jtstcur, 	// *  6  CURテスト
	jtstcur1, 	//    7  CURテスト 実行
	jmsr1,		// *  8  0～5A測定入力メッセージ
	jmsr2,		//    9  測定データ入力
	jmsrmsg,	// *  10 測定開始メッセージ 定電流出力オン
	jmsrlp,		// *  11 測定開始 定電流出力更新
	jmsrlp1,	//    12 定電流安定待ち
	jmsrlp2,	//    13 測定値出力
        jmsrbk,         // *  14 測定中断
        jsmpl1,         // *  15 サンプリングテスト
        jsmpl2,         //    16 サンプリングテスト電流値入力
        jsmpl3,         //    17 サンプリングテストD/A出力,開始
        jsmpl4,         //    18 サンプリングテスト結果表示
        jtstiad,        // *  19 内蔵A/Dテスト
        jmon,           // *  20 モニター
};

/******************************/
/*      セットアップ          */
/******************************/

/*****   セットアップ          *****/
//    ATmega328Pのレジスタを直接制御
void setup()
{
//  I/Oイニシャル
    PORTB = 0b00110011;
    DDRB  = 0b00111111;      // port指定
    //          |||||+----  PB0 o D/A CS
    //          ||||+-----  PB1 o D/A SCK
    //          |||+------  PB2 o D/A SDI
    //          ||+-------  PB3 o CC on/off Hでon
    //          |+--------  PB4 o (no use)
    //          +---------  PB5 o P-P RESET on/off Lでリセット
    PORTC = 0b00000000;
    DDRC  = 0b00111100;
    //          ||||++----  PC0 i A/D0 交流電圧
    //          ||||+-----  PC1 i A/D1 交流電流
    //          ++++------  PC  o TEST POINT
    DIDR0 = 0b00000011;     // デジタル禁止
    PORTD = 0b00001000;
    DDRD  = 0b10110100;
    //        |||||||+----  RXD
    //        ||||||+-----  TXD
    //        |||||+------  PD2 o A/D SCK
    //        ||||+-------  PD3 i A/D DOUT (pull up)
    //        |||+--------  PD4 o A/D DIN
    //        ||+---------  PD5 o A/D CS
    //        |+----------  PD6 i CC Err入力 
    //        +-----------  PD7 o A/D CS
//  A/D割り込み
    ADCSRA = 0b10001110;
    //         |||||+++---  ADPS 1/64 250kHz
    //         ||||+------  ADIE 割り込み許可
    //         |||+-------  ADIF
    //         ||+--------  ADATE
    //         |+---------  ADSC
    //         +----------  ADEN
//  タイマー0 ：1msタイマー
    TCCR0A = 0b00000010;
    //               ++----  WGM01,00 CTCモード
    TCCR0B = 0b00000011;
    //             |+++----  プリスケーラ1/64
    //             +-------  WGM02
    OCR0A  = 250-1;          // 16MHz / 64 / 250 = 1kHz
    TIMSK0 = 0b00000010;
    //               +-----  OCIE0A
//  タイマー2 ：A/D変換開始 
    TCCR2A = 0b00000010;
    //               ++----  WGM21,20 CTCモード
    TCCR2B = 0b00000011;
    //             |+++----  プリスケーラ1/32
    //             +-------  WGM22
//  OCR2A  = 75-1;          // 16MHz / 32 / 75 = 6.66kHz 150uS
    OCR2A  = 80-1;          // 16MHz / 32 / 80 = 6.25kHz 160uS
    TIMSK2 = 0b00000010;
    //               +-----  OCIE2A
//  シリアル
    Serial.begin(9600);     //  シリアル通信
    sei();                   // 割り込み開始
}

/******************************/
/*      ループ                */
/******************************/
void loop()
{
//  PC3_X;                    // (!!!) PC3 26pin ﾄｸﾞﾙ
    if(f_tm1ms){              // 1ms割り込み
        f_tm1ms = 0;
        adavr();              // A/D平均処理
    }
    rxbff();                  // シリアル1行受信
    tblexc[exc_j]();          // テーブル実行
}

/*====    end of "ACADPT2"    ====*/

